# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from skytorch.torch.server import service_pb2 as skytorch_dot_torch_dot_server_dot_service__pb2

GRPC_GENERATED_VERSION = '1.78.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in skytorch/torch/server/service_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ServiceStub(object):
    """Tensor management service for SkyTorch PyTorch backend
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.UpdateTensor = channel.stream_unary(
                '/skytorch.torch.Service/UpdateTensor',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
                _registered_method=True)
        self.GetTensor = channel.unary_stream(
                '/skytorch.torch.Service/GetTensor',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.GetTensorRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.FromString,
                _registered_method=True)
        self.CopyTensor = channel.unary_unary(
                '/skytorch.torch.Service/CopyTensor',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.CopyTensorRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
                _registered_method=True)
        self.DeleteTensors = channel.unary_unary(
                '/skytorch.torch.Service/DeleteTensors',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.DeleteTensorsRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
                _registered_method=True)
        self.ExecuteAtenOperation = channel.unary_unary(
                '/skytorch.torch.Service/ExecuteAtenOperation',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenResponse.FromString,
                _registered_method=True)
        self.StreamOperations = channel.stream_stream(
                '/skytorch.torch.Service/StreamOperations',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.StreamRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.StreamResponse.FromString,
                _registered_method=True)
        self.ExecuteFunction = channel.unary_stream(
                '/skytorch.torch.Service/ExecuteFunction',
                request_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionRequest.SerializeToString,
                response_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionEvent.FromString,
                _registered_method=True)


class ServiceServicer(object):
    """Tensor management service for SkyTorch PyTorch backend
    """

    def UpdateTensor(self, request_iterator, context):
        """Tensor lifecycle management
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CopyTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteTensors(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteAtenOperation(self, request, context):
        """ATen operation execution
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamOperations(self, request_iterator, context):
        """Bidirectional streaming for low-latency operations
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteFunction(self, request, context):
        """Execute a function on the server
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'UpdateTensor': grpc.stream_unary_rpc_method_handler(
                    servicer.UpdateTensor,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.SerializeToString,
            ),
            'GetTensor': grpc.unary_stream_rpc_method_handler(
                    servicer.GetTensor,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.GetTensorRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.SerializeToString,
            ),
            'CopyTensor': grpc.unary_unary_rpc_method_handler(
                    servicer.CopyTensor,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.CopyTensorRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.SerializeToString,
            ),
            'DeleteTensors': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteTensors,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.DeleteTensorsRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.SerializeToString,
            ),
            'ExecuteAtenOperation': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteAtenOperation,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenResponse.SerializeToString,
            ),
            'StreamOperations': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamOperations,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.StreamRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.StreamResponse.SerializeToString,
            ),
            'ExecuteFunction': grpc.unary_stream_rpc_method_handler(
                    servicer.ExecuteFunction,
                    request_deserializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionRequest.FromString,
                    response_serializer=skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionEvent.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'skytorch.torch.Service', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('skytorch.torch.Service', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Service(object):
    """Tensor management service for SkyTorch PyTorch backend
    """

    @staticmethod
    def UpdateTensor(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/skytorch.torch.Service/UpdateTensor',
            skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/skytorch.torch.Service/GetTensor',
            skytorch_dot_torch_dot_server_dot_service__pb2.GetTensorRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.TensorChunk.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CopyTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/skytorch.torch.Service/CopyTensor',
            skytorch_dot_torch_dot_server_dot_service__pb2.CopyTensorRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteTensors(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/skytorch.torch.Service/DeleteTensors',
            skytorch_dot_torch_dot_server_dot_service__pb2.DeleteTensorsRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.TensorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExecuteAtenOperation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/skytorch.torch.Service/ExecuteAtenOperation',
            skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteAtenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamOperations(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/skytorch.torch.Service/StreamOperations',
            skytorch_dot_torch_dot_server_dot_service__pb2.StreamRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.StreamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExecuteFunction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/skytorch.torch.Service/ExecuteFunction',
            skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionRequest.SerializeToString,
            skytorch_dot_torch_dot_server_dot_service__pb2.ExecuteFunctionEvent.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""
Base metrics source interface for modular metrics collection.
"""
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Type


# Static registry mapping source names to (module_path, class_name) tuples
# This allows dynamic importing of metrics sources based on CLI configuration
_METRICS_SOURCES_REGISTRY: Dict[str, tuple[str, str]] = {
    "nvidia-gpu": ("skytorch.server.nvidia.metrics", "NvidiaGPUMetricsSource"),
}


class MetricData:
    """
    Represents a single metric data point.
    """
    def __init__(
        self,
        name: str,
        value: float,
        *,
        metric_type: str = "GAUGE",
        unit: str = "",
        labels: Optional[Dict[str, str]] = None,
        timestamp: Optional[int] = None,
        help_text: str = ""
    ):
        """
        Initialize a metric data point.

        Args:
            name: Metric name (e.g., "gpu.temperature")
            value: Numeric value
            metric_type: Type of metric ("GAUGE", "COUNTER", "HISTOGRAM")
            unit: Unit of measurement (e.g., "celsius", "percent")
            labels: Dictionary of labels/tags
            timestamp: Unix timestamp in seconds (default: current time)
            help_text: Description of the metric
        """
        self.name = name
        self.value = value
        self.metric_type = metric_type
        self.unit = unit
        self.labels = labels or {}
        self.timestamp = timestamp or int(time.time())
        self.help_text = help_text


class MetricsSource(ABC):
    """
    Abstract base class for metrics sources.

    Metrics sources provide metrics from different backends (NVIDIA GPU, AMD GPU, CPU, etc.)
    """

    @abstractmethod
    def get_source_name(self) -> str:
        """
        Get the name/identifier of this metrics source.

        Returns:
            Source name (e.g., "nvidia-gpu", "amd-gpu")
        """
        pass

    @abstractmethod
    def is_available(self) -> bool:
        """
        Check if this metrics source is available.

        Returns:
            True if the source can provide metrics, False otherwise
        """
        pass

    @abstractmethod
    def collect(
        self,
        *,
        metric_names: Optional[List[str]] = None,
        label_filters: Optional[Dict[str, str]] = None
    ) -> List[MetricData]:
        """
        Collect current metrics from this source.

        Args:
            metric_names: Optional list of specific metric names to collect
            label_filters: Optional label filters to apply

        Returns:
            List of metric data points
        """
        pass

    def get_metadata(self) -> Dict[str, str]:
        """
        Get metadata about this metrics source.

        Returns:
            Dictionary of metadata (e.g., version info, capabilities)
        """
        return {}

    def cleanup(self) -> None:
        """
        Cleanup resources used by this metrics source.

        Called when the metrics service is shutting down.
        """
        pass


def load_metrics_source(name: str) -> Optional[MetricsSource]:
    """
    Dynamically load and instantiate a metrics source by name.

    Args:
        name: Name of the metrics source

    Returns:
        The instantiated metrics source, or None if not found or failed to load
    """
    import importlib
    import logging

    logger = logging.getLogger(__name__)

    if name not in _METRICS_SOURCES_REGISTRY:
        logger.warning(f"Metrics source '{name}' not found in registry")
        return None

    module_path, class_name = _METRICS_SOURCES_REGISTRY[name]

    try:
        # Dynamically import the module
        module = importlib.import_module(module_path)

        # Get the class from the module
        source_class = getattr(module, class_name)

        # Instantiate the metrics source
        source = source_class()

        logger.debug(f"Loaded metrics source '{name}' from {module_path}.{class_name}")
        return source

    except ImportError as e:
        logger.warning(f"Failed to import metrics source '{name}': {e}")
        return None
    except AttributeError as e:
        logger.error(f"Class '{class_name}' not found in module '{module_path}': {e}")
        return None
    except Exception as e:
        logger.error(f"Failed to instantiate metrics source '{name}': {e}")
        return None


def list_metrics_source_names() -> List[str]:
    """
    Get list of all registered metrics source names.

    Returns:
        List of source names
    """
    return list(_METRICS_SOURCES_REGISTRY.keys())



"""
Metrics service for sky with support for modular metrics sources.
"""
from skytorch.server.metrics.source import (
    MetricsSource,
    MetricData,
    load_metrics_source,
    list_metrics_source_names,
)
from skytorch.server.metrics.service import MetricsServicer

__all__ = [
    "MetricsSource",
    "MetricData",
    "MetricsServicer",
    "load_metrics_source",
    "list_metrics_source_names",
]



"""
SkyTorch ATen Dispatch - Meta tensor execution for shape inference.

This module provides the fallback mechanism for ATen operations on SkyTorch devices.
It uses meta tensors to infer output shapes without moving data, then creates
output tensors on the SkyTorch device and executes operations remotely.
"""

import logging
import time
from dataclasses import dataclass
from typing import Any

import torch
import torch._dynamo  # Pre-import to avoid circular import issues during debugging

from skytorch.torch.backend import _client
from skytorch.torch.backend._client import map_args_kwargs
from skytorch.torch.profiler import PROFILING_ENABLED

logger = logging.getLogger(__name__)

# C++ _compute_dispatch_context is ~12x faster at computing cache keys than
# the Python _build_cache_key_with_context, but it fills the gRPC streaming
# pipeline faster. This increases latency at synchronous scalar fetch points
# (_local_scalar_dense), resulting in a net regression for workloads with
# frequent sync points (e.g., LLM inference with model.generate).
try:
    from skytorch.torch.backend._C import _compute_dispatch_context
except ImportError:
    _compute_dispatch_context = None

try:
    from skytorch.torch.backend._C import (
        _clear_shape_cache,
        _dispatch_cached_aten,
        _populate_shape_cache,
    )
except ImportError:
    _dispatch_cached_aten = None
    _populate_shape_cache = None
    _clear_shape_cache = None

try:
    from skytorch.torch.backend._C import _set_submit_callback, _clear_submit_callback
except ImportError:
    _set_submit_callback = None
    _clear_submit_callback = None

try:
    from skytorch.torch.backend._C import _register_storage_tensor_mapping, _register_tensor_id
except ImportError:
    _register_tensor_id = None
    _register_storage_tensor_mapping = None

try:
    from skytorch.torch.backend._C import _set_submit_method
except ImportError:
    _set_submit_method = None

try:
    from skytorch.torch.backend._C import _increment_ops_counter
except ImportError:
    _increment_ops_counter = None

try:
    from skytorch.torch.backend._C import _set_python_fallback, _clear_python_fallback
except ImportError:
    _set_python_fallback = None
    _clear_python_fallback = None

try:
    from skytorch.torch.backend._C import _take_pending_fused_result
except ImportError:
    _take_pending_fused_result = None

try:
    from skytorch.torch.backend._C import _set_profiling_enabled
except ImportError:
    _set_profiling_enabled = None

if PROFILING_ENABLED and _set_profiling_enabled is not None:
    _set_profiling_enabled(True)

_submit_callback_registered = False

_SHAPE_CACHE_MAX_SIZE = 4096
_sky_device_cache: dict[int, torch.device] = {}

# Cached stream_manager per device index — stable after device creation.
# Eliminates device_manager -> remote_info -> compute -> _grpc_client.stream
# attribute chain on every dispatch (~3 dict/attribute lookups per op).
_cached_stream_managers: dict[int, object] = {}


@dataclass(frozen=True, slots=True)
class _OutputMeta:
    """Cached metadata for a single output tensor."""

    shape: tuple[int, ...]
    stride: tuple[int, ...]
    dtype: torch.dtype
    storage_offset: int
    # Index into input tensor list if this output aliases an input, else -1
    alias_input: int


# Cache: op signature -> list of output metadata
# Key is int (C++ 64-bit hash) when C++ extension available, else tuple (Python)
_shape_cache: dict = {}


_UNCACHEABLE = object()

# Cache for str(op) conversion — avoids ~1μs per call (called twice on cache hit)
_op_name_cache: dict = {}


def _get_op_name(op: torch._ops.OpOverload) -> str:
    try:
        return _op_name_cache[op]
    except KeyError:
        name = str(op)
        _op_name_cache[op] = name
        return name


def _build_cache_key_with_context(
    op: torch._ops.OpOverload,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> tuple[tuple | None, torch.device | None, list[torch.Tensor]]:
    """Build cache key, resolve sky device, and collect input tensors in one pass.

    Returns (cache_key, sky_device, input_tensors).
    cache_key is None if args contain unhashable types.
    sky_device is the first sky device found in args/kwargs.
    input_tensors is a list of unique sky tensors (by storage ptr).
    """
    parts: list = [_get_op_name(op)]
    sky_device: torch.device | None = None
    input_tensors: list[torch.Tensor] = []
    seen_ptrs: set[int] = set()

    # Track storage sharing: data_ptr -> group ID
    storage_groups: dict[int, int] = {}
    next_group = 0
    uncacheable = False

    def _process(obj: Any) -> Any:
        nonlocal sky_device, next_group, uncacheable
        if isinstance(obj, torch.Tensor):
            if obj.device.type == "sky":
                if sky_device is None:
                    sky_device = obj.device
                storage = obj.untyped_storage()
                ptr = storage.data_ptr()
                if ptr not in seen_ptrs:
                    seen_ptrs.add(ptr)
                    input_tensors.append(obj)
                if ptr not in storage_groups:
                    storage_groups[ptr] = next_group
                    next_group += 1
                return (
                    "T",
                    tuple(obj.shape),
                    obj.dtype,
                    tuple(obj.stride()),
                    obj.storage_offset(),
                    storage.nbytes(),
                    storage_groups[ptr],
                )
            # Non-sky tensor (e.g., cpu scalar) — still build key
            storage = obj.untyped_storage()
            ptr = storage.data_ptr()
            if ptr not in storage_groups:
                storage_groups[ptr] = next_group
                next_group += 1
            return (
                "T",
                tuple(obj.shape),
                obj.dtype,
                tuple(obj.stride()),
                obj.storage_offset(),
                storage.nbytes(),
                storage_groups[ptr],
            )
        if isinstance(obj, (list, tuple)):
            return tuple(_process(v) for v in obj)
        if isinstance(obj, torch.device):
            if obj.type == "sky" and sky_device is None:
                sky_device = obj
            return ("D", obj.type, obj.index)
        if obj is None:
            return None
        if isinstance(obj, (int, float, bool, str, torch.dtype, torch.memory_format, torch.layout)):
            return obj
        uncacheable = True
        return _UNCACHEABLE

    # Check kwargs device first (like _resolve_device did)
    if "device" in kwargs and isinstance(kwargs["device"], torch.device):
        if kwargs["device"].type == "sky":
            sky_device = kwargs["device"]

    for arg in args:
        k = _process(arg)
        if uncacheable:
            return None, sky_device, input_tensors
        parts.append(k)

    if kwargs:
        for key in sorted(kwargs):
            k = _process(kwargs[key])
            if uncacheable:
                return None, sky_device, input_tensors
            parts.append((key, k))

    return tuple(parts), sky_device, input_tensors


def _create_meta_tensor_from_sky(
    sky_tensor: torch.Tensor,
    meta_storage_cache: dict[torch.UntypedStorage, torch.UntypedStorage],
) -> torch.Tensor:
    """Create a meta tensor that mirrors a sky tensor, including storage sharing."""
    original_storage = sky_tensor.untyped_storage()

    # Create or reuse meta storage to preserve storage sharing relationships
    if original_storage not in meta_storage_cache:
        nbytes = original_storage.nbytes()
        meta_storage_cache[original_storage] = torch.UntypedStorage(nbytes, device="meta")

    meta_storage = meta_storage_cache[original_storage]

    # Create meta tensor with same metadata as sky tensor
    meta_tensor = torch.empty(0, dtype=sky_tensor.dtype, device="meta")
    meta_tensor.set_(
        meta_storage,
        sky_tensor.storage_offset(),
        sky_tensor.shape,
        sky_tensor.stride(),
    )

    return meta_tensor


def _execute_meta_operation(
    op: torch._ops.OpOverload,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    devices: list[torch.device],
) -> tuple[Any, dict, bool]:
    """Execute operation on meta tensors for shape inference and device resolution."""
    original_tensors: dict[torch.UntypedStorage, torch.Tensor] = {}
    meta_storage_cache: dict[torch.UntypedStorage, torch.UntypedStorage] = {}
    has_empty_cpu = False

    if "device" in kwargs:
        devices.append(kwargs["device"])

    def to_meta_tensor(obj):
        nonlocal has_empty_cpu

        if isinstance(obj, torch.Tensor) and obj.device.type == "sky":
            devices.append(obj.device)

        # Convert tensor to meta for shape inference
        if isinstance(obj, torch.Tensor):
            # Validate device type: must be sky or cpu scalar (0-dim)
            if obj.device.type != "sky":
                if obj.device.type == "cpu" and obj.dim() == 0:
                    # cpu scalar tensors are allowed - pass through as-is
                    return obj
                elif obj.device.type == "cpu" and obj.numel() == 0:
                    # empty cpu tensors are allowed - promote to meta for shape inference
                    has_empty_cpu = True
                    return torch.empty(obj.shape, dtype=obj.dtype, device="meta")
                else:
                    raise RuntimeError(
                        f"Cannot mix {obj.device.type} tensors with sky tensors "
                        f"in {op}. "
                        f"Got {obj.device.type} tensor with shape={list(obj.shape)}, "
                        f"dtype={obj.dtype}. "
                        f"Only 0-dimensional cpu scalar tensors are allowed. "
                        f"Please move your tensor to the sky device first."
                    )

            meta_tensor = _create_meta_tensor_from_sky(obj, meta_storage_cache)
            original_tensors[meta_tensor.untyped_storage()] = obj
            return meta_tensor

        # Convert device arguments to meta device
        if isinstance(obj, torch.device):
            return torch.device("meta")

        return obj

    meta_args, meta_kwargs = map_args_kwargs(to_meta_tensor, args, kwargs)
    meta_result = op(*meta_args, **meta_kwargs)

    return meta_result, original_tensors, has_empty_cpu


def _create_output_tensors(
    meta_outputs: list,
    original_tensors: dict[torch.UntypedStorage, torch.Tensor],
    sky_device: torch.device,
) -> list[torch.Tensor | None]:
    """Create output tensors based on meta execution results with proper alias detection."""
    output_tensors: list[torch.Tensor | None] = []

    for meta_output in meta_outputs:
        # Handle None outputs (common in backward operations)
        if meta_output is None:
            output_tensors.append(None)
            continue

        meta_storage = meta_output.untyped_storage()

        if meta_storage in original_tensors:
            # This output uses storage from an existing tensor (view/alias)
            original_tensor = original_tensors[meta_storage]

            # Resize if the original tensor is uninitialized and output has data
            if original_tensor.numel() == 0 and meta_output.numel() > 0:
                original_tensor.resize_(meta_output.shape)

            tensor = original_tensor.as_strided(
                meta_output.shape,
                meta_output.stride(),
                meta_output.storage_offset(),
            )
            output_tensors.append(tensor)
        else:
            # Create new tensor with new storage
            tensor = torch.empty_strided(
                meta_output.shape,
                meta_output.stride(),
                dtype=meta_output.dtype,
                device=sky_device,
            )
            # Record the storage mapping for future outputs that might alias
            original_tensors[meta_storage] = tensor
            output_tensors.append(tensor)

    return output_tensors


def _execute_with_static_outputs(
    op: torch._ops.OpOverload,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    sky_device: torch.device,
    meta_result: Any,
    original_tensors: dict[torch.UntypedStorage, torch.Tensor],
    cache_key: Any = None,
    input_tensors: list[torch.Tensor] | None = None,
    fused_cache_hash: int | None = None,
) -> Any:
    """Execute operation using meta tensors for shape inference."""
    # Normalize meta_result to list
    if isinstance(meta_result, torch.Tensor):
        meta_outputs = [meta_result]
    elif isinstance(meta_result, (tuple, list)):
        meta_outputs = list(meta_result)
    else:
        meta_outputs = []

    # Create output tensors based on meta shapes
    output_tensors = (
        _create_output_tensors(meta_outputs, original_tensors, sky_device) if meta_outputs else []
    )

    # Populate the shape cache
    if cache_key is not None and input_tensors is not None:
        _populate_cache(cache_key, meta_outputs, original_tensors, input_tensors)
        # Also populate the C++ shape cache for fused dispatch
        if fused_cache_hash is not None:
            _populate_cpp_cache(fused_cache_hash, meta_outputs, original_tensors, input_tensors)

    # Execute operation remotely via gRPC
    _client.execute_aten_operation(
        sky_device=sky_device,
        op_name=_get_op_name(op),
        args=args,
        kwargs=kwargs,
        output_tensors=output_tensors,
    )

    # Return results
    if len(output_tensors) > 1:
        return tuple(output_tensors)
    elif output_tensors:
        return output_tensors[0]
    else:
        return None


def _populate_cache(
    cache_key: Any,
    meta_outputs: list,
    original_tensors: dict[torch.UntypedStorage, torch.Tensor],
    input_tensors: list[torch.Tensor],
) -> None:
    """Store shape inference results in cache."""
    if len(_shape_cache) >= _SHAPE_CACHE_MAX_SIZE:
        return

    # Build storage -> input index mapping for alias detection
    storage_to_input: dict[int, int] = {}
    for i, t in enumerate(input_tensors):
        ptr = t.untyped_storage().data_ptr()
        if ptr not in storage_to_input:
            storage_to_input[ptr] = i

    cached_outputs: list[_OutputMeta | None] = []
    for meta_output in meta_outputs:
        if meta_output is None:
            cached_outputs.append(None)
            continue

        meta_storage = meta_output.untyped_storage()
        alias_input = -1

        if meta_storage in original_tensors:
            original = original_tensors[meta_storage]
            ptr = original.untyped_storage().data_ptr()
            alias_input = storage_to_input.get(ptr, -1)

        cached_outputs.append(
            _OutputMeta(
                shape=tuple(meta_output.shape),
                stride=tuple(meta_output.stride()),
                dtype=meta_output.dtype,
                storage_offset=meta_output.storage_offset(),
                alias_input=alias_input,
            )
        )

    _shape_cache[cache_key] = cached_outputs


def _create_outputs_from_cache(
    cached_outputs: list[_OutputMeta | None],
    input_tensors: list[torch.Tensor],
    sky_device: torch.device,
) -> list[torch.Tensor | None]:
    """Create output tensors from cached shape metadata."""
    output_tensors: list[torch.Tensor | None] = []

    for meta in cached_outputs:
        if meta is None:
            output_tensors.append(None)
            continue

        if meta.alias_input >= 0:
            # Output aliases an input tensor
            original = input_tensors[meta.alias_input]

            if original.numel() == 0 and len(meta.shape) > 0 and all(s > 0 for s in meta.shape):
                original.resize_(meta.shape)

            tensor = original.as_strided(meta.shape, meta.stride, meta.storage_offset)
            output_tensors.append(tensor)
        else:
            # New storage
            tensor = torch.empty_strided(
                meta.shape,
                meta.stride,
                dtype=meta.dtype,
                device=sky_device,
            )
            output_tensors.append(tensor)

    return output_tensors


def _find_offending_arg(
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> str | None:
    """Find the argument location of a non-sky, non-scalar tensor.

    Scans args and kwargs (recursing into lists/tuples) and returns a
    human-readable path like ``args[1]`` or ``kwargs['mask'][0]`` for
    the first tensor that is not on the sky device and not a 0-dim cpu
    scalar.  Returns ``None`` if no offending tensor is found.

    Only called on the error path so performance is not a concern.
    """

    def _scan(obj: Any, path: str) -> str | None:
        if isinstance(obj, torch.Tensor):
            if obj.device.type != "sky" and not (obj.device.type == "cpu" and obj.dim() == 0):
                return path
        elif isinstance(obj, (list, tuple)):
            for i, item in enumerate(obj):
                result = _scan(item, f"{path}[{i}]")
                if result is not None:
                    return result
        return None

    for i, arg in enumerate(args):
        result = _scan(arg, f"args[{i}]")
        if result is not None:
            return result

    for key, val in kwargs.items():
        result = _scan(val, f"kwargs['{key}']")
        if result is not None:
            return result

    return None


def _collect_input_tensors(args: tuple[Any, ...], kwargs: dict[str, Any]) -> list[torch.Tensor]:
    """Collect all sky tensors from args/kwargs in order (for cache alias tracking)."""
    tensors: list[torch.Tensor] = []
    seen: set[int] = set()

    def _collect(obj: Any) -> None:
        if isinstance(obj, torch.Tensor) and obj.device.type == "sky":
            ptr = obj.untyped_storage().data_ptr()
            if ptr not in seen:
                seen.add(ptr)
                tensors.append(obj)
        elif isinstance(obj, (list, tuple)):
            for v in obj:
                _collect(v)

    for arg in args:
        _collect(arg)
    for val in kwargs.values():
        _collect(val)
    return tensors


def _get_stream_manager(dev_idx: int):
    """Get cached stream manager for a device index."""
    sm = _cached_stream_managers.get(dev_idx)
    if sm is not None:
        return sm

    from skytorch.torch.backend._device import device_manager

    remote_info = device_manager._local_to_remote.get(dev_idx)
    if remote_info is None:
        raise RuntimeError(
            "No Compute context available for ATen operation. "
            "Ensure you are within an 'async with Compute(...):' block."
        )
    compute = remote_info.compute
    if compute._grpc_client is None:
        raise RuntimeError(
            f"Compute '{compute.name}' is not ready. The gRPC client has not been initialized."
        )

    sm = compute._grpc_client.stream
    _cached_stream_managers[dev_idx] = sm

    # Register C++ submit callback on first stream manager resolution
    global _submit_callback_registered
    if not _submit_callback_registered and _set_submit_callback is not None:
        _set_submit_callback(_submit_and_register)
        _submit_callback_registered = True

    # Set up C++ raw submit buffer for bypassing Python on the fast path
    sm._setup_cpp_submit()

    return sm


def _submit_and_register(
    raw_bytes: bytes,
    new_tensor_ids: list,
    new_storage_ids: list,
    new_storage_nbytes: list,
    dev_idx: int,
) -> None:
    """Submit serialized request and register new tensors locally."""
    from skytorch.torch.backend._storage import storage_manager

    if _increment_ops_counter is not None:
        _increment_ops_counter()
    else:
        from skytorch.torch.backend.aten import scalar as _scalar_mod

        _scalar_mod._ops_since_last_sync += 1

    stream_manager = _get_stream_manager(dev_idx)
    stream_manager.submit_execute_aten_bytes(raw_bytes)

    # Register new tensors locally and in C++ tracking set after successful submission
    for tensor_id, storage_id, nb in zip(
        new_tensor_ids, new_storage_ids, new_storage_nbytes, strict=True
    ):
        storage_manager.register_storage(
            storage_id=storage_id,
            nbytes=nb,
            device_index=dev_idx,
        )
        with storage_manager._lock:
            storage_manager._storage_to_tensors[storage_id].add(tensor_id)
        if _register_tensor_id is not None:
            _register_tensor_id(tensor_id)
        if _register_storage_tensor_mapping is not None:
            _register_storage_tensor_mapping(storage_id, tensor_id)


def _unpack_outputs(output_tensors: list) -> Any:
    """Unpack output tensors list into the expected return format."""
    if len(output_tensors) > 1:
        return tuple(output_tensors)
    elif output_tensors:
        return output_tensors[0]
    else:
        return None


def _populate_cpp_cache(
    cache_hash: int,
    meta_outputs: list,
    original_tensors: dict,
    input_tensors: list,
) -> None:
    """Populate the C++ shape cache alongside the Python cache."""
    if _populate_shape_cache is None:
        return

    # Build storage -> input index mapping for alias detection
    storage_to_input: dict[int, int] = {}
    for i, t in enumerate(input_tensors):
        ptr = t.untyped_storage().data_ptr()
        if ptr not in storage_to_input:
            storage_to_input[ptr] = i

    cpp_metas = []
    for meta_output in meta_outputs:
        if meta_output is None:
            cpp_metas.append(None)
            continue

        meta_storage = meta_output.untyped_storage()
        alias_input = -1

        if meta_storage in original_tensors:
            original = original_tensors[meta_storage]
            ptr = original.untyped_storage().data_ptr()
            alias_input = storage_to_input.get(ptr, -1)

        cpp_metas.append(
            (
                list(meta_output.shape),
                list(meta_output.stride()),
                _dtype_to_scalar_type_int(meta_output.dtype),
                meta_output.storage_offset(),
                alias_input,
            )
        )

    _populate_shape_cache(cache_hash, cpp_metas)


# Map torch.dtype to c10::ScalarType integer value
_DTYPE_TO_SCALAR_TYPE: dict[torch.dtype, int] = {
    torch.float32: 6,
    torch.float64: 7,
    torch.float16: 5,
    torch.bfloat16: 15,
    torch.int32: 3,
    torch.int64: 4,
    torch.int16: 2,
    torch.int8: 1,
    torch.uint8: 0,
    torch.bool: 11,
    torch.complex64: 9,
    torch.complex128: 10,
}


def _dtype_to_scalar_type_int(dtype: torch.dtype) -> int:
    """Convert torch.dtype to c10::ScalarType integer."""
    result = _DTYPE_TO_SCALAR_TYPE.get(dtype)
    if result is not None:
        return result
    # Fallback for newer dtypes
    if hasattr(torch, "float8_e5m2") and dtype == torch.float8_e5m2:
        return 23
    if hasattr(torch, "float8_e4m3fn") and dtype == torch.float8_e4m3fn:
        return 24
    raise ValueError(f"Unsupported dtype for cache: {dtype}")


def _sky_kernel_fallback(
    op: torch._ops.OpOverload,
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Execute PyTorch operations on SkyTorch devices using meta tensor dispatch."""
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"Dispatching {op}")

    try:
        op_name = _get_op_name(op)

        if PROFILING_ENABLED:
            from skytorch.torch.profiler import ClientProfiler

            _prof = ClientProfiler.get()
            _t0 = time.perf_counter_ns()
            # Inter-op gap is now measured in C++ fallback_kernel for both
            # the fast path and Python path, avoiding redundant perf_counter_ns
            # calls here. Only track first dispatch and total ops count.
            if _prof.first_dispatch_ns == 0:
                _prof.first_dispatch_ns = _t0
            _prof.total_ops += 1

        # Fused C++ path: hash + cache lookup + output creation + serialization
        # Only handles cache hits; misses/uncacheable fall through to existing path
        fused_cache_hash = None
        if _dispatch_cached_aten is not None:
            # Check if C++ fallback_kernel already called dispatch_cached_aten
            # and stored the result. This avoids the redundant call when the
            # Python fallback is invoked from FallbackKernel.cpp.
            pending = (
                _take_pending_fused_result() if _take_pending_fused_result is not None else None
            )
            if pending is not None:
                # Unwrap from (result,) wrapper — the result is the original
                # dispatch_cached_aten return value (None or Tuple(3))
                fused_result = pending[0]
            else:
                fused_result = _dispatch_cached_aten(op_name, args, kwargs)

            if fused_result is not None and len(fused_result) == 1:
                # Cache hit with C++ submit callback — already submitted
                # Note: _ops_since_last_sync already incremented by _submit_and_register callback

                if PROFILING_ENABLED:
                    _t1 = time.perf_counter_ns()
                    _prof.cache_key_build.add(_t1 - _t0)
                    _prof.execute_dispatch.add(0)
                    _prof.cache_hits += 1
                    _prof.last_dispatch_end = _t1
                    _prof.last_dispatch_ns = _t1

                return fused_result[0]

            if fused_result is not None and len(fused_result) == 6:
                # Cache hit without callback — Python-side submission
                (
                    unpacked_output,
                    raw_bytes,
                    new_tensor_ids,
                    new_storage_ids,
                    new_storage_nbytes,
                    dev_idx,
                ) = fused_result

                if PROFILING_ENABLED:
                    _t1 = time.perf_counter_ns()
                    _prof.cache_key_build.add(_t1 - _t0)
                    _prof.cache_hits += 1

                _submit_and_register(
                    raw_bytes, new_tensor_ids, new_storage_ids, new_storage_nbytes, dev_idx
                )

                if PROFILING_ENABLED:
                    _t3 = time.perf_counter_ns()
                    _prof.execute_dispatch.add(_t3 - _t1)
                    _prof.last_dispatch_end = _t3
                    _prof.last_dispatch_ns = _t3

                return unpacked_output

            if fused_result is not None and len(fused_result) == 3:
                # Cache miss — save hash so we can dual-populate C++ cache below
                fused_cache_hash = fused_result[0]

            # Fall through to existing path for cache misses and uncacheable

        # Compute cache key, resolve device, and collect tensors
        if _compute_dispatch_context is not None:
            # C++ fast path: single walk of args/kwargs
            cache_hash, input_tensors, sky_device_idx = _compute_dispatch_context(
                op_name, args, kwargs
            )
            cache_key = cache_hash if cache_hash != 0 else None
            if sky_device_idx >= 0:
                sky_device = _sky_device_cache.get(sky_device_idx)
                if sky_device is None:
                    sky_device = torch.device("sky", sky_device_idx)
                    _sky_device_cache[sky_device_idx] = sky_device
            else:
                sky_device = None
        else:
            # Python fallback
            cache_key, sky_device, input_tensors = _build_cache_key_with_context(op, args, kwargs)

        if PROFILING_ENABLED:
            _t1 = time.perf_counter_ns()
            _prof.cache_key_build.add(_t1 - _t0)

        # Try shape cache first
        if cache_key is not None:
            cached = _shape_cache.get(cache_key)
            if cached is not None and sky_device is not None:
                output_tensors = _create_outputs_from_cache(cached, input_tensors, sky_device)

                if PROFILING_ENABLED:
                    _t2 = time.perf_counter_ns()
                    _prof.output_creation.add(_t2 - _t1)
                    _prof.cache_hits += 1

                _client.execute_aten_operation(
                    sky_device=sky_device,
                    op_name=op_name,
                    args=args,
                    kwargs=kwargs,
                    output_tensors=output_tensors,
                )

                if PROFILING_ENABLED:
                    _t3 = time.perf_counter_ns()
                    _prof.execute_dispatch.add(_t3 - _t2)
                    _prof.last_dispatch_end = _t3
                    _prof.last_dispatch_ns = _t3

                if len(output_tensors) > 1:
                    return tuple(output_tensors)
                elif output_tensors:
                    return output_tensors[0]
                else:
                    return None

        if PROFILING_ENABLED:
            _prof.cache_misses += 1

        # Cache miss — full meta execution
        devices: list[torch.device] = []

        meta_result, original_tensors, has_empty_cpu = _execute_meta_operation(
            op, args, kwargs, devices
        )

        if not devices:
            raise RuntimeError(f"Could not determine sky device for operation {op}")

        sky_device = devices[0]

        if has_empty_cpu:

            def _promote(obj):
                if (
                    isinstance(obj, torch.Tensor)
                    and obj.device.type == "cpu"
                    and obj.numel() == 0
                    and obj.dim() > 0
                ):
                    return torch.empty(obj.shape, dtype=obj.dtype, device=sky_device)
                return obj

            args, kwargs = map_args_kwargs(_promote, args, kwargs)
            # Re-collect input tensors after promotion
            input_tensors = _collect_input_tensors(args, kwargs) if cache_key is not None else None
        else:
            input_tensors = input_tensors if cache_key is not None else None

        result = _execute_with_static_outputs(
            op,
            args,
            kwargs,
            sky_device,
            meta_result,
            original_tensors,
            cache_key=cache_key,
            input_tensors=input_tensors,
            fused_cache_hash=fused_cache_hash,
        )

        if PROFILING_ENABLED:
            _tend = time.perf_counter_ns()
            _prof.last_dispatch_end = _tend
            _prof.last_dispatch_ns = _tend

        return result
    except RuntimeError as e:
        if "Cannot mix" in str(e):
            location = _find_offending_arg(args, kwargs)
            if location:
                raise RuntimeError(f"{e} ({location})") from None
        raise
    except NotImplementedError:
        # Meta execution not implemented for this operation
        raise NotImplementedError(
            f"Operation {op} is not supported on sky device. " f"Meta tensor execution failed."
        )


# Register _sky_kernel_fallback as the Python fallback for the C++ boxed fallback kernel.
# This is called for cache misses and uncacheable ops from FallbackKernel.cpp.
if _set_python_fallback is not None:
    _set_python_fallback(_sky_kernel_fallback)



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1PodResourceClaim(BaseModel):
    """
    PodResourceClaim references exactly one ResourceClaim, either directly or by naming a ResourceClaimTemplate which is then turned into a ResourceClaim for the pod.  It adds a name to it that uniquely identifies the ResourceClaim inside the Pod. Containers that need access to the ResourceClaim reference it with this name.
    """ # noqa: E501
    name: StrictStr = Field(description="Name uniquely identifies this resource claim inside the pod. This must be a DNS_LABEL.")
    resource_claim_name: Optional[StrictStr] = Field(default=None, description="ResourceClaimName is the name of a ResourceClaim object in the same namespace as this pod.  Exactly one of ResourceClaimName and ResourceClaimTemplateName must be set.", alias="resourceClaimName")
    resource_claim_template_name: Optional[StrictStr] = Field(default=None, description="ResourceClaimTemplateName is the name of a ResourceClaimTemplate object in the same namespace as this pod.  The template will be used to create a new ResourceClaim, which will be bound to this pod. When this pod is deleted, the ResourceClaim will also be deleted. The pod name and resource name, along with a generated component, will be used to form a unique name for the ResourceClaim, which will be recorded in pod.status.resourceClaimStatuses.  This field is immutable and no changes will be made to the corresponding ResourceClaim by the control plane after creating the ResourceClaim.  Exactly one of ResourceClaimName and ResourceClaimTemplateName must be set.", alias="resourceClaimTemplateName")
    __properties: ClassVar[List[str]] = ["name", "resourceClaimName", "resourceClaimTemplateName"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PodResourceClaim from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PodResourceClaim from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "name": obj.get("name") if obj.get("name") is not None else '',
            "resourceClaimName": obj.get("resourceClaimName"),
            "resourceClaimTemplateName": obj.get("resourceClaimTemplateName")
        })
        return _obj





"""
SkyTorch ATen Dynamic Operations - Handlers for data-dependent output shapes.

This module provides implementations for ATen operations that have
data-dependent output shapes (e.g., masked_select, nonzero). These
operations cannot use meta tensor execution for shape inference and
require computing the actual output size at runtime.
"""

from typing import Any

import torch

from skytorch.torch.backend import _client


def _handle_masked_select(
        op: torch._ops.OpOverload,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
) -> Any:
    """Handle masked_select with data-dependent output shape."""
    # Extract self tensor and mask from args
    self_tensor = args[0]
    mask = args[1]

    # Compute output size by counting True values in mask
    # This executes on the sky server via the normal dispatch path
    output_size = int(mask.sum().cpu().item())

    # Create output tensor with the correct size
    output = torch.empty(
        output_size,
        dtype=self_tensor.dtype,
        device=self_tensor.device,
    )

    # Execute the operation with the pre-allocated output
    _client.execute_aten_operation(
        sky_device=self_tensor.device,
        op_name=str(op),
        args=args,
        kwargs=kwargs,
        output_tensors=[output],
    )

    return output


def _masked_select(self_tensor: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """Handle masked_select with data-dependent output shape."""
    return _handle_masked_select(
        torch.ops.aten.masked_select.default,
        (self_tensor, mask),
        {},
    )


def _masked_select_out(
    self_tensor: torch.Tensor, mask: torch.Tensor, out: torch.Tensor
) -> torch.Tensor:
    """Handle masked_select.out with data-dependent output shape."""
    return _handle_masked_select(
        torch.ops.aten.masked_select.out,
        (self_tensor, mask),
        {"out": out},
    )



"""
KV cache utilities for speculative scalar with HuggingFace Transformers.

Provides a DynamicCache subclass that supports cropping after speculation
overshoot, including for models with sliding window attention.
"""

from transformers.cache_utils import DynamicCache, DynamicSlidingWindowLayer


class _CropSafeLayer(DynamicSlidingWindowLayer):
    """DynamicSlidingWindowLayer that supports crop() after the window fills.

    The base class forbids crop() once cumulative_length >= sliding_window,
    but removing recent tail entries is always safe (evicted head entries
    aren't involved). Also computes masks from actual stored key count
    instead of the fixed ``sliding_window - 1`` formula — identical during
    normal operation (stored = sliding_window - 1), but correct after crop.
    """

    def crop(self, max_length):
        if max_length < 0:
            max_length = self.keys.shape[-2] + max_length
        stored = self.keys.shape[-2]
        if stored <= max_length:
            return
        trimmed = stored - max_length
        self.keys = self.keys[..., :max_length, :]
        self.values = self.values[..., :max_length, :]
        self.cumulative_length = max(0, self.cumulative_length - trimmed)

    def get_mask_sizes(self, cache_position):
        query_length = cache_position.shape[0]
        if not self.is_initialized:
            return query_length, 0
        stored = self.keys.shape[-2]
        kv_length = stored + query_length
        kv_offset = max(self.cumulative_length - self.sliding_window + 1, 0)
        return kv_length, kv_offset


class SpeculativeCache(DynamicCache):
    """DynamicCache with crop-safe sliding window layers.

    Replaces ``DynamicSlidingWindowLayer`` instances with ``_CropSafeLayer``
    so that speculation overshoot entries can be removed after generation
    via :meth:`crop_overshoot`.

    Usage::

        cache = SpeculativeCache(config=model.config)
        output = model.generate(..., past_key_values=cache)

        overshoot = ...  # number of tokens beyond the valid stopping point
        if overshoot > 0:
            cache.crop_overshoot(overshoot)
    """

    def __init__(self, config):
        super().__init__(config=config)
        for i, layer in enumerate(self.layers):
            if type(layer) is DynamicSlidingWindowLayer:
                safe = _CropSafeLayer(sliding_window=layer.sliding_window)
                self.layers[i] = safe

    def crop_overshoot(self, overshoot):
        """Remove the last ``overshoot`` entries from all cache layers."""
        for layer in self.layers:
            if not layer.is_initialized:
                continue
            stored = layer.keys.shape[-2]
            if overshoot > 0 and stored > overshoot:
                layer.crop(stored - overshoot)



"""
SkyTorch Server.

This package provides:
- serve: Async gRPC server for tensor operations
- Compute: Lightweight compute for local testing without Kubernetes
- compute: Decorator for automatic Compute lifecycle management
"""

from skytorch.torch.server.service import TensorServicer
from skytorch.torch.server.server import serve
from skytorch.torch.server.compute import Compute, compute

__all__ = ["serve", "Compute", "compute"]



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1PersistentVolumeClaimCondition(BaseModel):
    """
    PersistentVolumeClaimCondition contains details about state of pvc
    """ # noqa: E501
    last_probe_time: Optional[datetime] = Field(default=None, description="lastProbeTime is the time we probed the condition.", alias="lastProbeTime")
    last_transition_time: Optional[datetime] = Field(default=None, description="lastTransitionTime is the time the condition transitioned from one status to another.", alias="lastTransitionTime")
    message: Optional[StrictStr] = Field(default=None, description="message is the human-readable message indicating details about last transition.")
    reason: Optional[StrictStr] = Field(default=None, description="reason is a unique, this should be a short, machine understandable string that gives the reason for condition's last transition. If it reports \"Resizing\" that means the underlying persistent volume is being resized.")
    status: StrictStr = Field(description="Status is the status of the condition. Can be True, False, Unknown. More info: https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-claim-v1/#:~:text=state%20of%20pvc-,conditions.status,-(string)%2C%20required")
    type: StrictStr = Field(description="Type is the type of the condition. More info: https://kubernetes.io/docs/reference/kubernetes-api/config-and-storage-resources/persistent-volume-claim-v1/#:~:text=set%20to%20%27ResizeStarted%27.-,PersistentVolumeClaimCondition,-contains%20details%20about")
    __properties: ClassVar[List[str]] = ["lastProbeTime", "lastTransitionTime", "message", "reason", "status", "type"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PersistentVolumeClaimCondition from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PersistentVolumeClaimCondition from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "lastProbeTime": obj.get("lastProbeTime"),
            "lastTransitionTime": obj.get("lastTransitionTime"),
            "message": obj.get("message"),
            "reason": obj.get("reason"),
            "status": obj.get("status") if obj.get("status") is not None else '',
            "type": obj.get("type") if obj.get("type") is not None else ''
        })
        return _obj





"""
SkyTorch Deferred Scalar Tensor.

A torch.Tensor subclass that wraps a zero-filled CPU tensor and a
concurrent.futures.Future. Overrides tolist() and item() to block
on the future and fill data before delegating to the real tensor.

Used by async copy to make .cpu() non-blocking for single-element
tensors during LLM generation.
"""

from __future__ import annotations

from concurrent.futures import Future

import torch


class DeferredScalarTensor(torch.Tensor):
    """Tensor subclass that lazily resolves its scalar value from a Future.

    Created via _make_subclass to share storage with a real CPU tensor,
    so .shape, .device, .dtype, .dim() all work natively through C++.
    The actual value is resolved on first access via tolist() or item().

    Uses a class-level dict keyed by id(self) because _make_subclass
    instances don't support __slots__ or reliable per-instance __dict__.
    """

    _futures: dict[int, Future] = {}

    @staticmethod
    def create(shape: tuple | list, dtype: torch.dtype, future: Future) -> DeferredScalarTensor:
        """Create a tensor backed by a zero-filled CPU tensor.

        Args:
            shape: Shape of the tensor
            dtype: Data type of the tensor
            future: Future that resolves to the scalar value

        Returns:
            DeferredScalarTensor wrapping the backing tensor
        """
        backing = torch.zeros(shape, dtype=dtype)
        t = torch.Tensor._make_subclass(DeferredScalarTensor, backing)
        DeferredScalarTensor._futures[id(t)] = future
        return t

    def _resolve(self):
        """Block on the future and fill the backing tensor with the resolved value."""
        future = DeferredScalarTensor._futures.pop(id(self), None)
        if future is not None:
            self.fill_(future.result())

    def tolist(self):
        self._resolve()
        return super().tolist()

    def item(self):
        self._resolve()
        return super().item()

    def __del__(self):
        DeferredScalarTensor._futures.pop(id(self), None)



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictBytes, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from skytorch.client.models.io_k8s_apimachinery_pkg_apis_meta_v1_object_meta import IoK8sApimachineryPkgApisMetaV1ObjectMeta
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1Secret(BaseModel):
    """
    Secret holds secret data of a certain type. The total bytes of the values in the Data field must be less than MaxSecretSize bytes.
    """ # noqa: E501
    api_version: Optional[StrictStr] = Field(default=None, description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources", alias="apiVersion")
    data: Optional[Dict[str, Union[StrictBytes, StrictStr]]] = Field(default=None, description="Data contains the secret data. Each key must consist of alphanumeric characters, '-', '_' or '.'. The serialized form of the secret data is a base64 encoded string, representing the arbitrary (possibly non-string) data value here. Described in https://tools.ietf.org/html/rfc4648#section-4")
    immutable: Optional[StrictBool] = Field(default=None, description="Immutable, if set to true, ensures that data stored in the Secret cannot be updated (only object metadata can be modified). If not set to true, the field can be modified at any time. Defaulted to nil.")
    kind: Optional[StrictStr] = Field(default=None, description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds")
    metadata: Optional[IoK8sApimachineryPkgApisMetaV1ObjectMeta] = Field(default=None, description="Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata")
    string_data: Optional[Dict[str, StrictStr]] = Field(default=None, description="stringData allows specifying non-binary secret data in string form. It is provided as a write-only input field for convenience. All keys and values are merged into the data field on write, overwriting any existing values. The stringData field is never output when reading from the API.", alias="stringData")
    type: Optional[StrictStr] = Field(default=None, description="Used to facilitate programmatic handling of secret data. More info: https://kubernetes.io/docs/concepts/configuration/secret/#secret-types")
    __properties: ClassVar[List[str]] = ["apiVersion", "data", "immutable", "kind", "metadata", "stringData", "type"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1Secret from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict['metadata'] = self.metadata.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1Secret from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "apiVersion": obj.get("apiVersion"),
            "data": obj.get("data"),
            "immutable": obj.get("immutable"),
            "kind": obj.get("kind"),
            "metadata": IoK8sApimachineryPkgApisMetaV1ObjectMeta.from_dict(obj["metadata"]) if obj.get("metadata") is not None else None,
            "stringData": obj.get("stringData"),
            "type": obj.get("type")
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from skytorch.client.models.io_k8s_api_core_v1_container_status import IoK8sApiCoreV1ContainerStatus
from skytorch.client.models.io_k8s_api_core_v1_host_ip import IoK8sApiCoreV1HostIP
from skytorch.client.models.io_k8s_api_core_v1_pod_condition import IoK8sApiCoreV1PodCondition
from skytorch.client.models.io_k8s_api_core_v1_pod_extended_resource_claim_status import IoK8sApiCoreV1PodExtendedResourceClaimStatus
from skytorch.client.models.io_k8s_api_core_v1_pod_ip import IoK8sApiCoreV1PodIP
from skytorch.client.models.io_k8s_api_core_v1_pod_resource_claim_status import IoK8sApiCoreV1PodResourceClaimStatus
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1PodStatus(BaseModel):
    """
    PodStatus represents information about the status of a pod. Status may trail the actual state of a system, especially if the node that hosts the pod cannot contact the control plane.
    """ # noqa: E501
    conditions: Optional[List[IoK8sApiCoreV1PodCondition]] = Field(default=None, description="Current service state of pod. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#pod-conditions")
    container_statuses: Optional[List[IoK8sApiCoreV1ContainerStatus]] = Field(default=None, description="Statuses of containers in this pod. Each container in the pod should have at most one status in this list, and all statuses should be for containers in the pod. However this is not enforced. If a status for a non-existent container is present in the list, or the list has duplicate names, the behavior of various Kubernetes components is not defined and those statuses might be ignored. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#pod-and-container-status", alias="containerStatuses")
    ephemeral_container_statuses: Optional[List[IoK8sApiCoreV1ContainerStatus]] = Field(default=None, description="Statuses for any ephemeral containers that have run in this pod. Each ephemeral container in the pod should have at most one status in this list, and all statuses should be for containers in the pod. However this is not enforced. If a status for a non-existent container is present in the list, or the list has duplicate names, the behavior of various Kubernetes components is not defined and those statuses might be ignored. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#pod-and-container-status", alias="ephemeralContainerStatuses")
    extended_resource_claim_status: Optional[IoK8sApiCoreV1PodExtendedResourceClaimStatus] = Field(default=None, description="Status of extended resource claim backed by DRA.", alias="extendedResourceClaimStatus")
    host_ip: Optional[StrictStr] = Field(default=None, description="hostIP holds the IP address of the host to which the pod is assigned. Empty if the pod has not started yet. A pod can be assigned to a node that has a problem in kubelet which in turns mean that HostIP will not be updated even if there is a node is assigned to pod", alias="hostIP")
    host_ips: Optional[List[IoK8sApiCoreV1HostIP]] = Field(default=None, description="hostIPs holds the IP addresses allocated to the host. If this field is specified, the first entry must match the hostIP field. This list is empty if the pod has not started yet. A pod can be assigned to a node that has a problem in kubelet which in turns means that HostIPs will not be updated even if there is a node is assigned to this pod.", alias="hostIPs")
    init_container_statuses: Optional[List[IoK8sApiCoreV1ContainerStatus]] = Field(default=None, description="Statuses of init containers in this pod. The most recent successful non-restartable init container will have ready = true, the most recently started container will have startTime set. Each init container in the pod should have at most one status in this list, and all statuses should be for containers in the pod. However this is not enforced. If a status for a non-existent container is present in the list, or the list has duplicate names, the behavior of various Kubernetes components is not defined and those statuses might be ignored. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#pod-and-container-status", alias="initContainerStatuses")
    message: Optional[StrictStr] = Field(default=None, description="A human readable message indicating details about why the pod is in this condition.")
    nominated_node_name: Optional[StrictStr] = Field(default=None, description="nominatedNodeName is set only when this pod preempts other pods on the node, but it cannot be scheduled right away as preemption victims receive their graceful termination periods. This field does not guarantee that the pod will be scheduled on this node. Scheduler may decide to place the pod elsewhere if other nodes become available sooner. Scheduler may also decide to give the resources on this node to a higher priority pod that is created after preemption. As a result, this field may be different than PodSpec.nodeName when the pod is scheduled.", alias="nominatedNodeName")
    observed_generation: Optional[StrictInt] = Field(default=None, description="If set, this represents the .metadata.generation that the pod status was set based upon. This is an alpha field. Enable PodObservedGenerationTracking to be able to use this field.", alias="observedGeneration")
    phase: Optional[StrictStr] = Field(default=None, description="The phase of a Pod is a simple, high-level summary of where the Pod is in its lifecycle. The conditions array, the reason and message fields, and the individual container status arrays contain more detail about the pod's status. There are five possible phase values:  Pending: The pod has been accepted by the Kubernetes system, but one or more of the container images has not been created. This includes time before being scheduled as well as time spent downloading images over the network, which could take a while. Running: The pod has been bound to a node, and all of the containers have been created. At least one container is still running, or is in the process of starting or restarting. Succeeded: All containers in the pod have terminated in success, and will not be restarted. Failed: All containers in the pod have terminated, and at least one container has terminated in failure. The container either exited with non-zero status or was terminated by the system. Unknown: For some reason the state of the pod could not be obtained, typically due to an error in communicating with the host of the pod.  More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#pod-phase  Possible enum values:  - `\"Failed\"` means that all containers in the pod have terminated, and at least one container has terminated in a failure (exited with a non-zero exit code or was stopped by the system).  - `\"Pending\"` means the pod has been accepted by the system, but one or more of the containers has not been started. This includes time before being bound to a node, as well as time spent pulling images onto the host.  - `\"Running\"` means the pod has been bound to a node and all of the containers have been started. At least one container is still running or is in the process of being restarted.  - `\"Succeeded\"` means that all containers in the pod have voluntarily terminated with a container exit code of 0, and the system is not going to restart any of these containers.  - `\"Unknown\"` means that for some reason the state of the pod could not be obtained, typically due to an error in communicating with the host of the pod. Deprecated: It isn't being set since 2015 (74da3b14b0c0f658b3bb8d2def5094686d0e9095)")
    pod_ip: Optional[StrictStr] = Field(default=None, description="podIP address allocated to the pod. Routable at least within the cluster. Empty if not yet allocated.", alias="podIP")
    pod_ips: Optional[List[IoK8sApiCoreV1PodIP]] = Field(default=None, description="podIPs holds the IP addresses allocated to the pod. If this field is specified, the 0th entry must match the podIP field. Pods may be allocated at most 1 value for each of IPv4 and IPv6. This list is empty if no IPs have been allocated yet.", alias="podIPs")
    qos_class: Optional[StrictStr] = Field(default=None, description="The Quality of Service (QOS) classification assigned to the pod based on resource requirements See PodQOSClass type for available QOS classes More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-qos/#quality-of-service-classes  Possible enum values:  - `\"BestEffort\"` is the BestEffort qos class.  - `\"Burstable\"` is the Burstable qos class.  - `\"Guaranteed\"` is the Guaranteed qos class.", alias="qosClass")
    reason: Optional[StrictStr] = Field(default=None, description="A brief CamelCase message indicating details about why the pod is in this state. e.g. 'Evicted'")
    resize: Optional[StrictStr] = Field(default=None, description="Status of resources resize desired for pod's containers. It is empty if no resources resize is pending. Any changes to container resources will automatically set this to \"Proposed\" Deprecated: Resize status is moved to two pod conditions PodResizePending and PodResizeInProgress. PodResizePending will track states where the spec has been resized, but the Kubelet has not yet allocated the resources. PodResizeInProgress will track in-progress resizes, and should be present whenever allocated resources != acknowledged resources.")
    resource_claim_statuses: Optional[List[IoK8sApiCoreV1PodResourceClaimStatus]] = Field(default=None, description="Status of resource claims.", alias="resourceClaimStatuses")
    start_time: Optional[datetime] = Field(default=None, description="RFC 3339 date and time at which the object was acknowledged by the Kubelet. This is before the Kubelet pulled the container image(s) for the pod.", alias="startTime")
    __properties: ClassVar[List[str]] = ["conditions", "containerStatuses", "ephemeralContainerStatuses", "extendedResourceClaimStatus", "hostIP", "hostIPs", "initContainerStatuses", "message", "nominatedNodeName", "observedGeneration", "phase", "podIP", "podIPs", "qosClass", "reason", "resize", "resourceClaimStatuses", "startTime"]

    @field_validator('phase')
    def phase_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in set(['Failed', 'Pending', 'Running', 'Succeeded', 'Unknown']):
            raise ValueError("must be one of enum values ('Failed', 'Pending', 'Running', 'Succeeded', 'Unknown')")
        return value

    @field_validator('qos_class')
    def qos_class_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in set(['BestEffort', 'Burstable', 'Guaranteed']):
            raise ValueError("must be one of enum values ('BestEffort', 'Burstable', 'Guaranteed')")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PodStatus from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in conditions (list)
        _items = []
        if self.conditions:
            for _item_conditions in self.conditions:
                if _item_conditions:
                    _items.append(_item_conditions.to_dict())
            _dict['conditions'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in container_statuses (list)
        _items = []
        if self.container_statuses:
            for _item_container_statuses in self.container_statuses:
                if _item_container_statuses:
                    _items.append(_item_container_statuses.to_dict())
            _dict['containerStatuses'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in ephemeral_container_statuses (list)
        _items = []
        if self.ephemeral_container_statuses:
            for _item_ephemeral_container_statuses in self.ephemeral_container_statuses:
                if _item_ephemeral_container_statuses:
                    _items.append(_item_ephemeral_container_statuses.to_dict())
            _dict['ephemeralContainerStatuses'] = _items
        # override the default output from pydantic by calling `to_dict()` of extended_resource_claim_status
        if self.extended_resource_claim_status:
            _dict['extendedResourceClaimStatus'] = self.extended_resource_claim_status.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in host_ips (list)
        _items = []
        if self.host_ips:
            for _item_host_ips in self.host_ips:
                if _item_host_ips:
                    _items.append(_item_host_ips.to_dict())
            _dict['hostIPs'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in init_container_statuses (list)
        _items = []
        if self.init_container_statuses:
            for _item_init_container_statuses in self.init_container_statuses:
                if _item_init_container_statuses:
                    _items.append(_item_init_container_statuses.to_dict())
            _dict['initContainerStatuses'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in pod_ips (list)
        _items = []
        if self.pod_ips:
            for _item_pod_ips in self.pod_ips:
                if _item_pod_ips:
                    _items.append(_item_pod_ips.to_dict())
            _dict['podIPs'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in resource_claim_statuses (list)
        _items = []
        if self.resource_claim_statuses:
            for _item_resource_claim_statuses in self.resource_claim_statuses:
                if _item_resource_claim_statuses:
                    _items.append(_item_resource_claim_statuses.to_dict())
            _dict['resourceClaimStatuses'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1PodStatus from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "conditions": [IoK8sApiCoreV1PodCondition.from_dict(_item) for _item in obj["conditions"]] if obj.get("conditions") is not None else None,
            "containerStatuses": [IoK8sApiCoreV1ContainerStatus.from_dict(_item) for _item in obj["containerStatuses"]] if obj.get("containerStatuses") is not None else None,
            "ephemeralContainerStatuses": [IoK8sApiCoreV1ContainerStatus.from_dict(_item) for _item in obj["ephemeralContainerStatuses"]] if obj.get("ephemeralContainerStatuses") is not None else None,
            "extendedResourceClaimStatus": IoK8sApiCoreV1PodExtendedResourceClaimStatus.from_dict(obj["extendedResourceClaimStatus"]) if obj.get("extendedResourceClaimStatus") is not None else None,
            "hostIP": obj.get("hostIP"),
            "hostIPs": [IoK8sApiCoreV1HostIP.from_dict(_item) for _item in obj["hostIPs"]] if obj.get("hostIPs") is not None else None,
            "initContainerStatuses": [IoK8sApiCoreV1ContainerStatus.from_dict(_item) for _item in obj["initContainerStatuses"]] if obj.get("initContainerStatuses") is not None else None,
            "message": obj.get("message"),
            "nominatedNodeName": obj.get("nominatedNodeName"),
            "observedGeneration": obj.get("observedGeneration"),
            "phase": obj.get("phase"),
            "podIP": obj.get("podIP"),
            "podIPs": [IoK8sApiCoreV1PodIP.from_dict(_item) for _item in obj["podIPs"]] if obj.get("podIPs") is not None else None,
            "qosClass": obj.get("qosClass"),
            "reason": obj.get("reason"),
            "resize": obj.get("resize"),
            "resourceClaimStatuses": [IoK8sApiCoreV1PodResourceClaimStatus.from_dict(_item) for _item in obj["resourceClaimStatuses"]] if obj.get("resourceClaimStatuses") is not None else None,
            "startTime": obj.get("startTime")
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field
from typing import Any, ClassVar, Dict, List, Optional
from skytorch.client.models.io_k8s_api_core_v1_persistent_volume_claim_template import IoK8sApiCoreV1PersistentVolumeClaimTemplate
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1EphemeralVolumeSource(BaseModel):
    """
    Represents an ephemeral volume that is handled by a normal storage driver.
    """ # noqa: E501
    volume_claim_template: Optional[IoK8sApiCoreV1PersistentVolumeClaimTemplate] = Field(default=None, description="Will be used to create a stand-alone PVC to provision the volume. The pod in which this EphemeralVolumeSource is embedded will be the owner of the PVC, i.e. the PVC will be deleted together with the pod.  The name of the PVC will be `<pod name>-<volume name>` where `<volume name>` is the name from the `PodSpec.Volumes` array entry. Pod validation will reject the pod if the concatenated name is not valid for a PVC (for example, too long).  An existing PVC with that name that is not owned by the pod will *not* be used for the pod to avoid using an unrelated volume by mistake. Starting the pod is then blocked until the unrelated PVC is removed. If such a pre-created PVC is meant to be used by the pod, the PVC has to updated with an owner reference to the pod once the pod exists. Normally this should not be necessary, but it may be useful when manually reconstructing a broken cluster.  This field is read-only and no changes will be made by Kubernetes to the PVC after it has been created.  Required, must not be nil.", alias="volumeClaimTemplate")
    __properties: ClassVar[List[str]] = ["volumeClaimTemplate"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1EphemeralVolumeSource from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of volume_claim_template
        if self.volume_claim_template:
            _dict['volumeClaimTemplate'] = self.volume_claim_template.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1EphemeralVolumeSource from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "volumeClaimTemplate": IoK8sApiCoreV1PersistentVolumeClaimTemplate.from_dict(obj["volumeClaimTemplate"]) if obj.get("volumeClaimTemplate") is not None else None
        })
        return _obj





"""
SkyTorch profiling instrumentation.

Lightweight profiling gated by SKYTORCH_PROFILE=1 environment variable.
Overhead: ~200ns/op when enabled (~12ms total for 58K ops), zero-cost when disabled.
"""

import os
import sys

PROFILING_ENABLED = os.environ.get("SKYTORCH_PROFILE", "0") == "1"

# When set, profiler also writes to this file (useful when stderr is captured
# by a TUI alternate screen buffer, e.g. Textual).
_PROFILE_FILE = os.environ.get("SKYTORCH_PROFILE_FILE", "")


def _profile_output(text: str) -> None:
    """Write profiler output to stderr and optionally to a file."""
    sys.stderr.write(text)
    sys.stderr.flush()
    if _PROFILE_FILE:
        with open(_PROFILE_FILE, "a") as f:
            f.write(text)
            f.flush()


class Counter:
    """Accumulates timing and count data."""

    __slots__ = ("total_ns", "count", "max_ns")

    def __init__(self):
        self.total_ns: int = 0
        self.count: int = 0
        self.max_ns: int = 0

    def add(self, ns: int) -> None:
        self.total_ns += ns
        self.count += 1
        if ns > self.max_ns:
            self.max_ns = ns

    def add_count(self, n: int = 1) -> None:
        self.count += n

    @property
    def avg_us(self) -> float:
        return (self.total_ns / self.count / 1000) if self.count else 0.0

    @property
    def total_ms(self) -> float:
        return self.total_ns / 1_000_000

    @property
    def max_ms(self) -> float:
        return self.max_ns / 1_000_000

    @property
    def avg_ms(self) -> float:
        return (self.total_ns / self.count / 1_000_000) if self.count else 0.0


class ClientProfiler:
    """Singleton profiler for client-side dispatch breakdown."""

    _instance = None

    def __init__(self):
        # Dispatch counters
        self.cache_key_build = Counter()
        self.output_creation = Counter()
        self.execute_dispatch = Counter()
        self.cpp_serialization = Counter()
        self.event_loop_submit = Counter()

        # Sync counters
        self.sync_total = Counter()

        # Sync phase breakdown
        self.sync_flush = Counter()
        self.sync_wait = Counter()

        # Sync wait decomposition (server-provided timing)
        self.sync_network_rtt = Counter()
        self.sync_server_backlog = Counter()
        self.sync_server_handle = Counter()

        # Scalar speculation stats
        self.scalar_speculative_hits: int = 0
        self.scalar_speculative_misses: int = 0

        # Sync buffer state (non-timing accumulators)
        self.sync_mt_ops_total: int = 0
        self.sync_queue_depth_total: int = 0
        self.sync_queue_depth_max: int = 0

        # Batch counters
        self.batch_count = Counter()
        self.batch_size_total: int = 0
        self.batch_size_max: int = 0

        # Inter-op gap
        self.inter_op_gap = Counter()
        self.last_dispatch_end: int = 0

        # Cache stats
        self.cache_hits: int = 0
        self.cache_misses: int = 0
        self.total_ops: int = 0

        # Wall time
        self.first_dispatch_ns: int = 0
        self.last_dispatch_ns: int = 0

    @classmethod
    def get(cls) -> "ClientProfiler":
        if cls._instance is None:
            cls._instance = ClientProfiler()
        return cls._instance

    @classmethod
    def reset(cls) -> None:
        cls._instance = None

    def print_summary(self) -> None:
        # Fetch C++ fast path counters
        try:
            from skytorch.torch.backend._C import _get_cpp_profile_counters

            cpp = _get_cpp_profile_counters()
        except (ImportError, AttributeError):
            cpp = None

        cpp_ops = cpp["fast_path_count"][1] if cpp else 0
        total = self.total_ops + cpp_ops

        if total == 0:
            return

        wall_ms = (self.last_dispatch_ns - self.first_dispatch_ns) / 1_000_000
        sync_pct = (self.sync_total.total_ms / wall_ms * 100) if wall_ms > 0 else 0

        lines = [
            "",
            "=== SkyTorch Client Profile ===",
            f"Ops: {total:,} ({self.cache_hits:,} cache hits, "
            f"{self.cache_misses:,} misses"
            + (f", {cpp_ops:,} C++ fast path" if cpp_ops > 0 else "")
            + ")",
            f"Syncs: {self.sync_total.count:,}",
            "",
            "Dispatch (per-op avg / total):",
            f"  Cache key build:    {self.cache_key_build.avg_us:6.1f} us  |  "
            f"{self.cache_key_build.total_ms:,.0f} ms",
            f"  Output creation:    {self.output_creation.avg_us:6.1f} us  |  "
            f"{self.output_creation.total_ms:,.0f} ms",
            f"  Execute dispatch:   {self.execute_dispatch.avg_us:6.1f} us  |  "
            f"{self.execute_dispatch.total_ms:,.0f} ms",
            f"    C++ serialize:    {self.cpp_serialization.avg_us:6.1f} us  |  "
            f"{self.cpp_serialization.total_ms:,.0f} ms",
            f"    Event loop sub:   {self.event_loop_submit.avg_us:6.1f} us  |  "
            f"{self.event_loop_submit.total_ms:,.0f} ms",
            f"  Inter-op gap:       {self.inter_op_gap.avg_us:6.1f} us  |  "
            f"{self.inter_op_gap.total_ms:,.0f} ms",
            "",
        ]

        if cpp and cpp_ops > 0:
            _iv_ns, _ = cpp["ivalue_to_py_ns"]
            _dc_ns, _ = cpp["dispatch_cached_ns"]
            _sm_ns, _ = cpp["submit_ns"]
            _rw_ns, _ = cpp["rewrite_stack_ns"]
            _total_ns = _iv_ns + _dc_ns + _rw_ns
            _pct = cpp_ops / total * 100 if total > 0 else 0
            lines.extend(
                [
                    "C++ fast path:",
                    f"  Ops handled:        {cpp_ops:>8,} / {total:,} ({_pct:.1f}%)",
                    f"  IValue -> py args:  {_iv_ns / cpp_ops / 1000:6.2f} us  |  "
                    f"{_iv_ns / 1_000_000:,.0f} ms",
                    f"  dispatch_cached:    {_dc_ns / cpp_ops / 1000:6.2f} us  |  "
                    f"{_dc_ns / 1_000_000:,.0f} ms",
                    f"    submit:           {_sm_ns / cpp_ops / 1000:6.2f} us  |  "
                    f"{_sm_ns / 1_000_000:,.0f} ms",
                    f"  Rewrite stack:      {_rw_ns / cpp_ops / 1000:6.2f} us  |  "
                    f"{_rw_ns / 1_000_000:,.0f} ms",
                    f"  Total per-op:       {_total_ns / cpp_ops / 1000:6.2f} us  |  "
                    f"{_total_ns / 1_000_000:,.0f} ms",
                    "",
                ]
            )

        # Python fallback path and inter-op gap (from C++ counters)
        if cpp:
            _, _py_count = cpp.get("python_fallback_count", (0, 0))
            _py_ns, _ = cpp.get("python_fallback_ns", (0, 0))
            _gap_ns, _gap_count = cpp.get("inter_op_gap_ns", (0, 0))
            if _py_count > 0:
                lines.extend(
                    [
                        "Python fallback path:",
                        f"  Ops handled:        {_py_count:>8,} / {total:,} "
                        f"({_py_count / total * 100:.1f}%)",
                        f"  Total time:         {_py_ns / _py_count / 1000:6.1f} us  |  "
                        f"{_py_ns / 1_000_000:,.0f} ms",
                        "",
                    ]
                )
            if _gap_count > 0:
                _gap_ms = _gap_ns / 1_000_000
                _gap_pct = (_gap_ms / wall_ms * 100) if wall_ms > 0 else 0
                lines.extend(
                    [
                        "Inter-op gap (C++):",
                        f"  Total:              {_gap_ms:,.0f} ms ({_gap_pct:.0f}% of wall time)",
                        f"  Per-gap avg:        {_gap_ns / _gap_count / 1000:.1f} us  "
                        f"({_gap_count:,} gaps)",
                    ]
                )
                # Gap decomposition
                _gil_wait_ns, _ = cpp.get("gil_wait_ns", (0, 0))
                _ag_ns, _ = cpp.get("autograd_overhead_ns", (0, 0))
                _outside_ag_ns, _outside_ag_count = cpp.get(
                    "outside_autograd_ns", (0, 0)
                )
                _total_ops = cpp_ops + _py_count
                if _total_ops > 0:
                    # autograd_overhead includes fallback_kernel time inside it.
                    # "inside autograd overhead" = autograd_total - fallback_kernel_inner
                    # This is: guard setup + callBoxed dispatch + device guard + guard teardown
                    _fast_ns, _ = cpp.get("dispatch_cached_ns", (0, 0))
                    _iv_ns, _ = cpp.get("ivalue_to_py_ns", (0, 0))
                    _rw_ns, _ = cpp.get("rewrite_stack_ns", (0, 0))
                    _fallback_ns = _fast_ns + _iv_ns + _rw_ns + _py_ns
                    _ag_inner_ns = max(0, _ag_ns - _fallback_ns)
                    # "outside autograd" = Python model code + PyTorch outer
                    # dispatcher (Python binding boxing/unboxing, dispatch key
                    # selection). Measured directly between autograd exit → entry.
                    # "inside autograd overhead" = gap - outside - GIL
                    # This captures: boxed dispatch infra inside autograd
                    # (guard, callBoxed, device guard)
                    lines.extend(
                        [
                            f"  Decomposition:",
                            f"    Outside autograd: {_outside_ag_ns / 1_000_000:,.0f} ms  "
                            f"({_outside_ag_ns / max(1, _outside_ag_count) / 1000:.1f} us/op)"
                            f"  [Python model + outer dispatch]",
                            f"    Inside autograd:  {_ag_inner_ns / 1_000_000:,.0f} ms  "
                            f"({_ag_inner_ns / _total_ops / 1000:.1f} us/op)"
                            f"  [guard + callBoxed + device guard]",
                            f"    GIL acquisition:  {_gil_wait_ns / 1_000_000:,.0f} ms  "
                            f"({_gil_wait_ns / _total_ops / 1000:.1f} us/op)",
                        ]
                    )
                # Large gap decomposition (>1ms)
                _lg_gil_ns, _lg_count = cpp.get("large_gap_gil_ns", (0, 0))
                _lg_other_ns, _ = cpp.get("large_gap_other_ns", (0, 0))
                if _lg_count > 0:
                    _lg_total = _lg_gil_ns + _lg_other_ns
                    lines.extend(
                        [
                            f"  Large gaps (>1ms):  {_lg_count:,} gaps, "
                            f"{_lg_total / 1_000_000:,.0f} ms total",
                            f"    GIL contention:   {_lg_gil_ns / 1_000_000:,.0f} ms  "
                            f"({_lg_gil_ns / _lg_count / 1_000_000:.1f} ms avg)"
                            f"  [event loop holding GIL]",
                            f"    Other (pre-GIL):  {_lg_other_ns / 1_000_000:,.0f} ms  "
                            f"({_lg_other_ns / _lg_count / 1_000_000:.1f} ms avg)"
                            f"  [sync wait, Python, dispatch]",
                        ]
                    )
                # Gap histogram
                _hist = cpp.get("gap_histogram", None)
                if _hist and any(v > 0 for v in _hist):
                    _labels = ["<1us", "1-10us", "10-100us", "100us-1ms", "1-10ms", ">10ms"]
                    _parts = []
                    for lbl, cnt in zip(_labels, _hist):
                        if cnt > 0:
                            _parts.append(f"{lbl}:{cnt:,}")
                    lines.append(f"  Histogram:          {' | '.join(_parts)}")
                lines.append("")
            _, _gil_count = cpp.get("gil_release_count", (0, 0))
            if _gil_count > 0:
                lines.extend(
                    [
                        "GIL releases:",
                        f"  Count:              {_gil_count:,}",
                        "",
                    ]
                )

        lines.extend(
            [
                "Sync points:",
                f"  Total wait:         {self.sync_total.total_ms:,.0f} ms "
                f"({sync_pct:.0f}% of wall time)",
                f"  Avg / Max:          {self.sync_total.avg_ms:.1f} ms / "
                f"{self.sync_total.max_ms:.1f} ms",
                "",
            ]
        )

        if self.sync_flush.count > 0:
            _sync_count = self.sync_flush.count
            _avg_ops = self.sync_mt_ops_total / _sync_count
            _avg_qdepth = self.sync_queue_depth_total / _sync_count
            lines.extend(
                [
                    "Sync phases (per-sync avg / total):",
                    f"  Flush (buffers->queue): {self.sync_flush.avg_ms:.2f} ms  |  "
                    f"{self.sync_flush.total_ms:,.0f} ms",
                    f"  Wait (enqueue->result): {self.sync_wait.avg_ms:.2f} ms  |  "
                    f"{self.sync_wait.total_ms:,.0f} ms",
                    f"  Ops drained at sync:   {_avg_ops:.0f} avg",
                    f"  Queue depth at sync:   {_avg_qdepth:.1f} avg  |  "
                    f"max {self.sync_queue_depth_max}",
                    "",
                ]
            )

        if self.sync_network_rtt.count > 0:
            lines.extend(
                [
                    "Sync wait decomposition (per-sync avg / total):",
                    f"  Network RTT:          {self.sync_network_rtt.avg_ms:.2f} ms  |  "
                    f"{self.sync_network_rtt.total_ms:,.0f} ms",
                    f"  Server backlog:       {self.sync_server_backlog.avg_ms:.2f} ms  |  "
                    f"{self.sync_server_backlog.total_ms:,.0f} ms",
                    f"  Server handle (GPU):  {self.sync_server_handle.avg_ms:.2f} ms  |  "
                    f"{self.sync_server_handle.total_ms:,.0f} ms",
                    "",
                ]
            )

        if self.scalar_speculative_hits > 0 or self.scalar_speculative_misses > 0:
            _total_spec = self.scalar_speculative_hits + self.scalar_speculative_misses
            _hit_pct = (self.scalar_speculative_hits / _total_spec * 100) if _total_spec else 0
            lines.extend(
                [
                    "Scalar speculation:",
                    f"  Hits: {self.scalar_speculative_hits:,}  |  "
                    f"Misses: {self.scalar_speculative_misses:,}  |  "
                    f"Rate: {_hit_pct:.0f}%",
                    "",
                ]
            )

        lines.extend(
            [
                "Batching:",
                (
                    f"  Batches: {self.batch_count.count:,}  |  "
                    f"Avg size: {(self.batch_size_total / self.batch_count.count):.1f}"
                    if self.batch_count.count > 0
                    else "  Batches: 0"
                ),
                f"  Max: {self.batch_size_max}" if self.batch_count.count > 0 else "",
                "",
                f"Wall time: {wall_ms:,.0f} ms",
                "",
            ]
        )
        _profile_output("\n".join(lines))


class ServerProfiler:
    """Per-stream profiler for server-side processing."""

    def __init__(self):
        # Execution counters
        self.raw_execute = Counter()
        self.raw_batched_execute = Counter()
        self.total_ops: int = 0

        # Mixed batch tracking
        self.mixed_batch_calls: int = 0
        self.mixed_batch_ops: int = 0

        # Sync counters
        self.sync_handle = Counter()
        self.scalar_gpu_sync = Counter()
        self.scalar_lookup = Counter()

        # Idle time
        self.idle_time = Counter()
        # Hot idle: idle time only between consecutive execution ops (not model loading gaps)
        self.hot_idle = Counter()

        # Per-sync-cycle breakdown
        self.sync_backlog_ops = Counter()
        self.sync_backlog_time = Counter()
        self.sync_idle_before = Counter()
        self.sync_cycle_count: int = 0

        # Wall time
        self.stream_start_ns: int = 0
        self.stream_end_ns: int = 0

    def print_summary(self) -> None:
        if self.stream_start_ns == 0:
            return

        wall_ms = (self.stream_end_ns - self.stream_start_ns) / 1_000_000
        if wall_ms <= 0:
            return

        fire_and_forget = self.raw_execute.count + self.raw_batched_execute.count
        sync_count = self.sync_handle.count

        exec_ms = self.raw_execute.total_ms + self.raw_batched_execute.total_ms
        exec_pct = (exec_ms / wall_ms * 100) if wall_ms > 0 else 0
        sync_pct = (self.sync_handle.total_ms / wall_ms * 100) if wall_ms > 0 else 0
        idle_pct = (self.idle_time.total_ms / wall_ms * 100) if wall_ms > 0 else 0

        lines = [
            "",
            "=== SkyTorch Server Profile ===",
            f"Requests: {fire_and_forget + sync_count:,} "
            f"({fire_and_forget:,} fire-and-forget, {sync_count:,} sync)",
            f"Ops: {self.total_ops:,}",
            "",
            "Execution:",
            f"  Raw execute:        {self.raw_execute.count:,} calls | "
            f"{self.raw_execute.total_ms:,.0f} ms",
            f"  Batched execute:    {self.raw_batched_execute.count:,} calls | "
            f"{self.raw_batched_execute.total_ms:,.0f} ms"
            + (
                f"  ({self.mixed_batch_calls:,} mixed, "
                f"{self.mixed_batch_ops:,} ops)"
                if self.mixed_batch_calls > 0
                else ""
            ),
            f"  Total execution:    {exec_ms:,.0f} ms ({exec_pct:.0f}%)",
            f"  Sync handle:        {sync_count:,} calls | "
            f"{self.sync_handle.total_ms:,.0f} ms ({sync_pct:.0f}%)",
        ]

        if self.scalar_gpu_sync.count > 0:
            lines.append(
                f"    GPU sync (item):  {self.scalar_gpu_sync.avg_ms:.1f} ms avg | "
                f"{self.scalar_gpu_sync.total_ms:,.0f} ms total"
            )
        if self.scalar_lookup.count > 0:
            lines.append(
                f"    Tensor lookup:    {self.scalar_lookup.avg_us:.1f} us avg | "
                f"{self.scalar_lookup.total_ms:,.0f} ms total"
            )

        if self.sync_cycle_count > 0:
            _sc = self.sync_cycle_count
            lines.extend(
                [
                    "",
                    f"Sync cycles: {_sc}",
                    f"  Backlog ops/cycle:  " f"{(self.sync_backlog_ops.count / _sc):.0f} avg",
                    f"  Backlog exec time:  "
                    f"{self.sync_backlog_time.avg_ms:.2f} ms avg  |  "
                    f"{self.sync_backlog_time.total_ms:,.0f} ms total",
                    f"  Server idle before: "
                    f"{self.sync_idle_before.avg_ms:.2f} ms avg  |  "
                    f"{self.sync_idle_before.total_ms:,.0f} ms total",
                    f"  Scalar handle:      "
                    f"{self.sync_handle.avg_ms:.2f} ms avg  |  "
                    f"{self.sync_handle.total_ms:,.0f} ms total",
                ]
            )

        # C++ per-phase execution breakdown (Step 1)
        try:
            from skytorch.torch.server._C import (
                get_server_profile_counters,
                reset_server_profile_counters,
            )

            cpp_prof = get_server_profile_counters()
            _cpp_op_count = cpp_prof["op_count"]
            if _cpp_op_count > 0:
                _parse_ns = cpp_prof["parse_ns"]
                _callboxed_ns = cpp_prof["callboxed_ns"]
                _output_ns = cpp_prof["output_ns"]
                _total_profiled_ns = _parse_ns + _callboxed_ns + _output_ns
                lines.extend(
                    [
                        "",
                        f"C++ per-phase breakdown ({_cpp_op_count:,} ops):",
                        f"  Parse (header+args): "
                        f"{_parse_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                        f"{_parse_ns / 1_000_000:,.0f} ms",
                        f"  callBoxed dispatch:  "
                        f"{_callboxed_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                        f"{_callboxed_ns / 1_000_000:,.0f} ms",
                        f"  Output extraction:   "
                        f"{_output_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                        f"{_output_ns / 1_000_000:,.0f} ms",
                        f"  Total per-op:        "
                        f"{_total_profiled_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                        f"{_total_profiled_ns / 1_000_000:,.0f} ms",
                    ]
                )
                # Sub-phase parse breakdown
                _parse_meta_ns = cpp_prof.get("parse_meta_ns", 0)
                _parse_args_ns = cpp_prof.get("parse_args_ns", 0)
                _tensor_arg_count = cpp_prof.get("tensor_arg_count", 0)
                _metadata_create_count = cpp_prof.get("metadata_create_count", 0)
                if _parse_meta_ns > 0 or _parse_args_ns > 0:
                    lines.extend(
                        [
                            f"  Parse sub-phases:",
                            f"    Meta (hdr+hash+meta+OpInfo): "
                            f"{_parse_meta_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                            f"{_parse_meta_ns / 1_000_000:,.0f} ms",
                            f"    Args (parse+defaults+coerce):"
                            f" {_parse_args_ns / _cpp_op_count / 1000:6.1f} us/op  |  "
                            f"{_parse_args_ns / 1_000_000:,.0f} ms",
                        ]
                    )
                    if _tensor_arg_count > 0:
                        _per_tensor_arg_ns = _parse_args_ns / _tensor_arg_count
                        lines.append(
                            f"    Tensor args: {_tensor_arg_count:,} "
                            f"({_tensor_arg_count / _cpp_op_count:.1f}/op, "
                            f"{_per_tensor_arg_ns / 1000:.2f} us/tensor-arg)"
                        )
                    if _metadata_create_count > 0:
                        lines.append(
                            f"    Metadata creates: {_metadata_create_count:,} "
                            f"(direct TensorImpl)"
                        )
                # Python fallback instrumentation
                _kwargs_fb = cpp_prof.get("kwargs_fallback_count", 0)
                _blocked_fb = cpp_prof.get("blocked_fallback_count", 0)
                _coercion_fb = cpp_prof.get("coercion_fallback_count", 0)
                _fallback_exec_ns = cpp_prof.get("fallback_exec_ns", 0)
                _fb_total = _kwargs_fb + _blocked_fb + _coercion_fb
                if _fb_total > 0:
                    _fb_exec_us = (
                        _fallback_exec_ns / _fb_total / 1000 if _fb_total > 0 else 0
                    )
                    _kwargs_resolved = cpp_prof.get("kwargs_resolved_count", 0)
                    lines.extend(
                        [
                            f"  Python fallback ops: {_fb_total:,}",
                            f"    kwargs:            {_kwargs_fb:,}"
                            + (
                                f"  (resolved to callBoxed: {_kwargs_resolved:,})"
                                if _kwargs_resolved > 0
                                else ""
                            ),
                            f"    callboxed_blocked: {_blocked_fb:,}",
                            f"    coercion failure:  {_coercion_fb:,}",
                            f"    Execution time:    {_fb_exec_us:6.1f} us/op  |  "
                            f"{_fallback_exec_ns / 1_000_000:,.0f} ms",
                        ]
                    )
                    # Show blocked op names and error messages
                    _blocked_ops = cpp_prof.get("blocked_op_counts", {})
                    _blocked_errors = cpp_prof.get("blocked_op_errors", {})
                    if _blocked_ops:
                        _sorted = sorted(
                            _blocked_ops.items(), key=lambda x: x[1], reverse=True
                        )
                        lines.append(f"    Blocked ops (top {min(10, len(_sorted))}):")
                        for _op_name, _count in _sorted[:10]:
                            _err = _blocked_errors.get(_op_name, "")
                            _err_sfx = f"  [{_err[:120]}]" if _err else ""
                            lines.append(f"      {_op_name}: {_count:,}{_err_sfx}")
                # Copy_tensor inline profiling
                _ct_count = cpp_prof.get("copy_tensor_count", 0)
                if _ct_count > 0:
                    _ct_total = cpp_prof.get("copy_tensor_total_ns", 0)
                    _ct_gil = cpp_prof.get("copy_tensor_gil_ns", 0)
                    _ct_proto = cpp_prof.get("copy_tensor_proto_ns", 0)
                    _ct_meta = cpp_prof.get("copy_tensor_meta_ns", 0)
                    _ct_copy = cpp_prof.get("copy_tensor_copy_ns", 0)
                    _ct_meta_count = cpp_prof.get("copy_tensor_meta_count", 0)
                    lines.extend(
                        [
                            f"  Copy_tensor ops:     {_ct_count:,}  "
                            f"({_ct_total / _ct_count / 1000:.1f} us/op  |  "
                            f"{_ct_total / 1_000_000:,.0f} ms)",
                            f"    GIL acquire:       "
                            f"{_ct_gil / _ct_count / 1000:6.1f} us/op  |  "
                            f"{_ct_gil / 1_000_000:,.0f} ms",
                            f"    Protobuf parse:    "
                            f"{_ct_proto / _ct_count / 1000:6.1f} us/op  |  "
                            f"{_ct_proto / 1_000_000:,.0f} ms",
                            f"    Metadata create:   "
                            f"{_ct_meta / _ct_count / 1000:6.1f} us/op  |  "
                            f"{_ct_meta / 1_000_000:,.0f} ms"
                            + (
                                f"  ({_ct_meta_count:,} tensors)"
                                if _ct_meta_count > 0
                                else ""
                            ),
                            f"    copy_() exec:      "
                            f"{_ct_copy / _ct_count / 1000:6.1f} us/op  |  "
                            f"{_ct_copy / 1_000_000:,.0f} ms",
                        ]
                    )
                # Batch-level decomposition
                _bt_count = cpp_prof.get("batch_count", 0)
                _bt_loop = cpp_prof.get("batch_loop_ns", 0)
                _bt_boundary = cpp_prof.get("batch_boundary_ns", 0)
                _bt_op_wall = cpp_prof.get("batch_op_wall_ns", 0)
                _bt_interop = cpp_prof.get("batch_interop_ns", 0)
                if _bt_count > 0 and _bt_boundary > 0:
                    _bt_boundary_only = _bt_boundary - _bt_loop
                    _bt_op_internal = (
                        _parse_ns + _callboxed_ns + _output_ns + _fallback_exec_ns
                    )
                    _bt_op_overhead = _bt_op_wall - _bt_op_internal
                    _batched_exec_ms = self.raw_batched_execute.total_ms
                    _py_vs_cpp = _batched_exec_ms - _bt_boundary / 1_000_000
                    lines.extend(
                        [
                            "",
                            f"  Batch decomposition ({_bt_count:,} C++ batch calls):",
                            f"    C++ func wall:     "
                            f"{_bt_boundary / 1_000_000:,.0f} ms  "
                            f"(loop: {_bt_loop / 1_000_000:,.0f} ms  |  "
                            f"boundary: {_bt_boundary_only / 1_000_000:,.0f} ms)",
                            f"    Op wall (outside):  "
                            f"{_bt_op_wall / 1_000_000:,.0f} ms  "
                            f"(internal: {_bt_op_internal / 1_000_000:,.0f} ms  |  "
                            f"overhead: {_bt_op_overhead / 1_000_000:,.0f} ms)",
                            f"    Inter-op gap:       "
                            f"{_bt_interop / 1_000_000:,.0f} ms  "
                            f"({_bt_interop / _cpp_op_count / 1000:.2f} us/op)",
                            f"    Python↔C++ gap:     "
                            f"{_py_vs_cpp:,.0f} ms  "
                            f"(Python wall - C++ func wall)",
                        ]
                    )
            reset_server_profile_counters()
        except (ImportError, AttributeError):
            pass

        lines.extend(
            [
                "",
                f"  Idle (between req): {self.idle_time.total_ms:,.0f} ms ({idle_pct:.0f}%)",
                f"  Hot idle (exec→exec): {self.hot_idle.total_ms:,.0f} ms "
                f"({self.hot_idle.count:,} gaps, "
                f"{self.hot_idle.avg_us:.0f} us avg, "
                f"{self.hot_idle.max_ms:.1f} ms max)",
                "",
                f"Wall time: {wall_ms:,.0f} ms",
                "",
            ]
        )
        _profile_output("\n".join(lines))



# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field
from typing import Any, ClassVar, Dict, List, Optional
from skytorch.client.models.io_k8s_api_core_v1_endpoint_address import IoK8sApiCoreV1EndpointAddress
from skytorch.client.models.io_k8s_api_core_v1_endpoint_port import IoK8sApiCoreV1EndpointPort
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1EndpointSubset(BaseModel):
    """
    EndpointSubset is a group of addresses with a common set of ports. The expanded set of endpoints is the Cartesian product of Addresses x Ports. For example, given:   {    Addresses: [{\"ip\": \"10.10.1.1\"}, {\"ip\": \"10.10.2.2\"}],    Ports:     [{\"name\": \"a\", \"port\": 8675}, {\"name\": \"b\", \"port\": 309}]  }  The resulting set of endpoints can be viewed as:   a: [ 10.10.1.1:8675, 10.10.2.2:8675 ],  b: [ 10.10.1.1:309, 10.10.2.2:309 ]  Deprecated: This API is deprecated in v1.33+.
    """ # noqa: E501
    addresses: Optional[List[IoK8sApiCoreV1EndpointAddress]] = Field(default=None, description="IP addresses which offer the related ports that are marked as ready. These endpoints should be considered safe for load balancers and clients to utilize.")
    not_ready_addresses: Optional[List[IoK8sApiCoreV1EndpointAddress]] = Field(default=None, description="IP addresses which offer the related ports but are not currently marked as ready because they have not yet finished starting, have recently failed a readiness check, or have recently failed a liveness check.", alias="notReadyAddresses")
    ports: Optional[List[IoK8sApiCoreV1EndpointPort]] = Field(default=None, description="Port numbers available on the related IP addresses.")
    __properties: ClassVar[List[str]] = ["addresses", "notReadyAddresses", "ports"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1EndpointSubset from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in addresses (list)
        _items = []
        if self.addresses:
            for _item_addresses in self.addresses:
                if _item_addresses:
                    _items.append(_item_addresses.to_dict())
            _dict['addresses'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in not_ready_addresses (list)
        _items = []
        if self.not_ready_addresses:
            for _item_not_ready_addresses in self.not_ready_addresses:
                if _item_not_ready_addresses:
                    _items.append(_item_not_ready_addresses.to_dict())
            _dict['notReadyAddresses'] = _items
        # override the default output from pydantic by calling `to_dict()` of each item in ports (list)
        _items = []
        if self.ports:
            for _item_ports in self.ports:
                if _item_ports:
                    _items.append(_item_ports.to_dict())
            _dict['ports'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1EndpointSubset from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "addresses": [IoK8sApiCoreV1EndpointAddress.from_dict(_item) for _item in obj["addresses"]] if obj.get("addresses") is not None else None,
            "notReadyAddresses": [IoK8sApiCoreV1EndpointAddress.from_dict(_item) for _item in obj["notReadyAddresses"]] if obj.get("notReadyAddresses") is not None else None,
            "ports": [IoK8sApiCoreV1EndpointPort.from_dict(_item) for _item in obj["ports"]] if obj.get("ports") is not None else None
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApimachineryPkgApisMetaV1ListOptions(BaseModel):
    """
    ListOptions is the query options to a standard REST list call.
    """ # noqa: E501
    allow_watch_bookmarks: Optional[StrictBool] = Field(default=None, description="allowWatchBookmarks requests watch events with type \"BOOKMARK\". Servers that do not implement bookmarks may ignore this flag and bookmarks are sent at the server's discretion. Clients should not assume bookmarks are returned at any specific interval, nor may they assume the server will send any BOOKMARK event during a session. If this is not a watch, this field is ignored.", alias="allowWatchBookmarks")
    api_version: Optional[StrictStr] = Field(default=None, description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources", alias="apiVersion")
    var_continue: Optional[StrictStr] = Field(default=None, description="The continue option should be set when retrieving more results from the server. Since this value is server defined, clients may only use the continue value from a previous query result with identical query parameters (except for the value of continue) and the server may reject a continue value it does not recognize. If the specified continue value is no longer valid whether due to expiration (generally five to fifteen minutes) or a configuration change on the server, the server will respond with a 410 ResourceExpired error together with a continue token. If the client needs a consistent list, it must restart their list without the continue field. Otherwise, the client may send another list request with the token received with the 410 error, the server will respond with a list starting from the next key, but from the latest snapshot, which is inconsistent from the previous list results - objects that are created, modified, or deleted after the first list request will be included in the response, as long as their keys are after the \"next key\".  This field is not supported when watch is true. Clients may start a watch from the last resourceVersion value returned by the server and not miss any modifications.", alias="continue")
    field_selector: Optional[StrictStr] = Field(default=None, description="A selector to restrict the list of returned objects by their fields. Defaults to everything.", alias="fieldSelector")
    kind: Optional[StrictStr] = Field(default=None, description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds")
    label_selector: Optional[StrictStr] = Field(default=None, description="A selector to restrict the list of returned objects by their labels. Defaults to everything.", alias="labelSelector")
    limit: Optional[StrictInt] = Field(default=None, description="limit is a maximum number of responses to return for a list call. If more items exist, the server will set the `continue` field on the list metadata to a value that can be used with the same initial query to retrieve the next set of results. Setting a limit may return fewer than the requested amount of items (up to zero items) in the event all requested objects are filtered out and clients should only use the presence of the continue field to determine whether more results are available. Servers may choose not to support the limit argument and will return all of the available results. If limit is specified and the continue field is empty, clients may assume that no more results are available. This field is not supported if watch is true.  The server guarantees that the objects returned when using continue will be identical to issuing a single list call without a limit - that is, no objects created, modified, or deleted after the first request is issued will be included in any subsequent continued requests. This is sometimes referred to as a consistent snapshot, and ensures that a client that is using limit to receive smaller chunks of a very large result can ensure they see all possible objects. If objects are updated during a chunked list the version of the object that was present at the time the first list result was calculated is returned.")
    resource_version: Optional[StrictStr] = Field(default=None, description="resourceVersion sets a constraint on what resource versions a request may be served from. See https://kubernetes.io/docs/reference/using-api/api-concepts/#resource-versions for details.  Defaults to unset", alias="resourceVersion")
    resource_version_match: Optional[StrictStr] = Field(default=None, description="resourceVersionMatch determines how resourceVersion is applied to list calls. It is highly recommended that resourceVersionMatch be set for list calls where resourceVersion is set See https://kubernetes.io/docs/reference/using-api/api-concepts/#resource-versions for details.  Defaults to unset", alias="resourceVersionMatch")
    send_initial_events: Optional[StrictBool] = Field(default=None, description="`sendInitialEvents=true` may be set together with `watch=true`. In that case, the watch stream will begin with synthetic events to produce the current state of objects in the collection. Once all such events have been sent, a synthetic \"Bookmark\" event  will be sent. The bookmark will report the ResourceVersion (RV) corresponding to the set of objects, and be marked with `\"k8s.io/initial-events-end\": \"true\"` annotation. Afterwards, the watch stream will proceed as usual, sending watch events corresponding to changes (subsequent to the RV) to objects watched.  When `sendInitialEvents` option is set, we require `resourceVersionMatch` option to also be set. The semantic of the watch request is as following: - `resourceVersionMatch` = NotOlderThan   is interpreted as \"data at least as new as the provided `resourceVersion`\"   and the bookmark event is send when the state is synced   to a `resourceVersion` at least as fresh as the one provided by the ListOptions.   If `resourceVersion` is unset, this is interpreted as \"consistent read\" and the   bookmark event is send when the state is synced at least to the moment   when request started being processed. - `resourceVersionMatch` set to any other value or unset   Invalid error is returned.  Defaults to true if `resourceVersion=\"\"` or `resourceVersion=\"0\"` (for backward compatibility reasons) and to false otherwise.", alias="sendInitialEvents")
    timeout_seconds: Optional[StrictInt] = Field(default=None, description="Timeout for the list/watch call. This limits the duration of the call, regardless of any activity or inactivity.", alias="timeoutSeconds")
    watch: Optional[StrictBool] = Field(default=None, description="Watch for changes to the described resources and return them as a stream of add, update, and remove notifications. Specify resourceVersion.")
    __properties: ClassVar[List[str]] = ["allowWatchBookmarks", "apiVersion", "continue", "fieldSelector", "kind", "labelSelector", "limit", "resourceVersion", "resourceVersionMatch", "sendInitialEvents", "timeoutSeconds", "watch"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1ListOptions from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApimachineryPkgApisMetaV1ListOptions from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "allowWatchBookmarks": obj.get("allowWatchBookmarks"),
            "apiVersion": obj.get("apiVersion"),
            "continue": obj.get("continue"),
            "fieldSelector": obj.get("fieldSelector"),
            "kind": obj.get("kind"),
            "labelSelector": obj.get("labelSelector"),
            "limit": obj.get("limit"),
            "resourceVersion": obj.get("resourceVersion"),
            "resourceVersionMatch": obj.get("resourceVersionMatch"),
            "sendInitialEvents": obj.get("sendInitialEvents"),
            "timeoutSeconds": obj.get("timeoutSeconds"),
            "watch": obj.get("watch")
        })
        return _obj





# coding: utf-8

"""
    SkyTorch OpenAPI Spec

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: unversioned
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self

class IoK8sApiCoreV1FileKeySelector(BaseModel):
    """
    FileKeySelector selects a key of the env file.
    """ # noqa: E501
    key: StrictStr = Field(description="The key within the env file. An invalid key will prevent the pod from starting. The keys defined within a source may consist of any printable ASCII characters except '='. During Alpha stage of the EnvFiles feature gate, the key size is limited to 128 characters.")
    optional: Optional[StrictBool] = Field(default=False, description="Specify whether the file or its key must be defined. If the file or key does not exist, then the env var is not published. If optional is set to true and the specified key does not exist, the environment variable will not be set in the Pod's containers.  If optional is set to false and the specified key does not exist, an error will be returned during Pod creation.")
    path: StrictStr = Field(description="The path within the volume from which to select the file. Must be relative and may not contain the '..' path or start with '..'.")
    volume_name: StrictStr = Field(description="The name of the volume mount containing the env file.", alias="volumeName")
    __properties: ClassVar[List[str]] = ["key", "optional", "path", "volumeName"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1FileKeySelector from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of IoK8sApiCoreV1FileKeySelector from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "key": obj.get("key") if obj.get("key") is not None else '',
            "optional": obj.get("optional") if obj.get("optional") is not None else False,
            "path": obj.get("path") if obj.get("path") is not None else '',
            "volumeName": obj.get("volumeName") if obj.get("volumeName") is not None else ''
        })
        return _obj


